@@ -242,7 +242,37 @@ class GameNode(object):
         main variation.
         """
         node = self.add_variation(move, comment=comment)
-        self.promote_to_main(move)
+        self.variations.remove(node)
+        self.variations.insert(0, node)
+        return node
+
+    def main_line(self):
+        """Yields the moves of the main line starting in this node."""
+        node = self
+        while node.variations:
+            node = node.variations[0]
+            yield node.move
+
+    def add_line(self, moves, comment="", starting_comment="", nags=()):
+        """
+        Creates a sequence of child nodes for the given list of moves.
+        Adds *comment* and *nags* to the last node of the line and returns it.
+        """
+        node = self
+
+        # Add line.
+        for move in moves:
+            node = node.add_variation(move, starting_comment=starting_comment)
+            starting_comment = ""
+
+        # Merge comment and NAGs.
+        if node.comment:
+            node.comment +=  " " + comment
+        else:
+            node.comment = comment
+
+        node.nags.update(nags)
+
         return node
 
     def accept(self, visitor, _board=None):
@@ -407,7 +437,7 @@ class Game(GameNode):
 
         super(Game, self).accept(visitor, _board=self.board())
 
-        visitor.visit_result(self.headers["Result"])
+        visitor.visit_result(self.headers.get("Result", "*"))
         visitor.end_game()
         return visitor.result()
 
@@ -558,6 +588,7 @@ class GameModelCreator(BaseVisitor):
 
     def handle_error(self, error):
         LOGGER.exception("error during pgn parsing")
+        self.game.errors.append(error)
 
     def result(self):
         """
